from .basic_callbacks import Callback, CancelTrainException
from ..hook import Hooks
from ..utils import ToolBox as tbox
import torch
import torch.nn as nn
import math
from functools import partial
import matplotlib.pyplot as plt


class LR_Find(Callback):
    _order = 1

    def __init__(self, max_iter=100, min_lr=1e-6, max_lr=10, **kwargs):
        super().__init__(**kwargs)
        self.max_iter, self.min_lr, self.max_lr = max_iter, min_lr, max_lr
        self.best_loss = 1e9

    def begin_batch(self):
        if self.state != 'train':
            return
        pos = self.n_iter/self.max_iter
        lr = self.min_lr * (self.max_lr/self.min_lr) ** pos
        for pg in self.opt.param_groups:
            pg['lr'] = lr

    def after_step(self):
        if self.n_iter >= self.max_iter or self.loss > self.best_loss*10:
            raise CancelTrainException()
        if self.loss < self.best_loss:
            self.best_loss = self.loss


class ParamScheduler(Callback):
    _order = 1

    def __init__(self, pname, sched_funcs, **kwargs):
        super().__init__(**kwargs)
        self.pname, self.sched_funcs = pname, sched_funcs

    def begin_fit(self):
        if not isinstance(self.sched_funcs, (list, tuple)):
            self.sched_funcs = [self.sched_funcs] * len(self.opt.param_groups)

    def set_param(self):
        assert len(self.opt.param_groups) == len(self.sched_funcs)
        for pg, f in zip(self.opt.param_groups, self.sched_funcs):
            pg[self.pname] = f(self.p_epochs/self.epochs)

    def begin_batch(self):
        if self.state == 'train':
            self.set_param()


def annealer(f):
    def _inner(start, end):
        return partial(f, start, end)
    return _inner


@annealer
def sched_linear(start, end, pos):
    return start + pos*(end-start)


@annealer
def sched_cos(start, end, pos):
    return start + (1 + math.cos(math.pi*(1-pos))) * (end-start) / 2


@annealer
def sched_no(start, end, pos):  # pylint: disable=unused-argument
    return start


@annealer
def sched_exp(start, end, pos):
    return start * (end/start) ** pos


def cos_1cycle_anneal(start, high, end):
    return [sched_cos(start, high), sched_cos(high, end)]  # pylint: disable=no-value-for-parameter


def combine_scheds(pcts, scheds):
    assert sum(pcts) == 1.
    pcts = torch.tensor([0] + tbox.listify(pcts))
    assert torch.all(pcts >= 0)
    pcts = torch.cumsum(pcts, 0)

    def _inner(pos):
        idx = (pos >= pcts).nonzero().max()
        actual_pos = (pos-pcts[idx]) / (pcts[idx+1]-pcts[idx])
        return scheds[idx](actual_pos)
    return _inner


class LayerAnalysisCallback(Callback):
    _order = 3

    def __init__(self, forward=True, hist_span=10, **kwargs):
        """
        forward: True, 各层forward输出分析; False, 各层backward输出梯度分析
        hist_span: histgram的分析范围。前向分析时应当大一些，例如10；后向分析时应当小一些，例如0.1。
        """
        super().__init__(**kwargs)
        self.forward = forward
        self.hist_span = [-hist_span, hist_span]

    def begin_fit(self):
        def append_stats(hook, module, inputs, outputs):  # pylint: disable=unused-argument
            if not hasattr(hook, 'stats'):
                hook.stats = ([], [], [])
            if isinstance(outputs, tuple):  # backward hook
                outputs = outputs[0]
            means, stds, hists = hook.stats
            means.append(outputs[0].data.mean().cpu())
            stds .append(outputs[0].data.std().cpu())
            hists.append(outputs[0].data.cpu().histc(40, *self.hist_span))

        self.hooks = Hooks(self.model, append_stats, self.forward)

    def after_fit(self):
        self.hooks.remove()
        self.plot()

    def plot(self):
        mode = 'FORWARD' if self.forward else 'BACKWARD'
        # 均值与标准差
        fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(12, 4))
        for h in self.hooks:
            ms, ss, _ = h.stats
            ax0.plot(ms, label=h.name)
            ax1.plot(ss, label=h.name)
        ax0.legend(prop={'size': 6})
        ax0.set_title(f"{mode}: Mean", fontsize=16)
        ax1.legend(prop={'size': 6})
        ax1.set_title(f"{mode}: Standard deviation", fontsize=16)

        # 各层输出值的分布
        figsize = (15, int(len(self.hooks)*0.7))
        fig, axes = plt.subplots(int(math.ceil(len(self.hooks)/3)), 3, figsize=figsize)
        [ax.axis('off') for ax in axes.flatten()]  # pylint:disable=expression-not-assigned
        for ax, h in zip(axes.flatten(), self.hooks):
            ax.axis('on')
            hist_matrix = torch.stack(h.stats[2]).t().float().log1p()
            extent = [0, hist_matrix.size()[1], *self.hist_span]
            im = ax.imshow(hist_matrix, origin='lower', extent=extent, aspect='auto')
            ax.set_title(h.name)
            fig.colorbar(im, ax=ax, shrink=1.0)
        fig.subplots_adjust(hspace=0.6, top=1-0.75/figsize[1])
        fig.suptitle(f'{mode}: Histogram of values by "log(1+x)"', fontsize=16)
        # plt.tight_layout()

        # 各层输出值中接近0的值的比例
        figsize = (15, int(len(self.hooks)*0.7))
        fig, axes = plt.subplots(int(math.ceil(len(self.hooks)/3)), 3, figsize=figsize)
        [ax.axis('off') for ax in axes.flatten()]  # pylint:disable=expression-not-assigned
        for ax, h in zip(axes.flatten(), self.hooks):
            ax.axis('on')
            hist_matrix = torch.stack(h.stats[2]).t().float()
            tiny_ratio = hist_matrix[19:22].sum(0)/hist_matrix.sum(0)
            ax.plot(tiny_ratio)
            ax.set_ylim(0, 1.02)
            ax.set_title(h.name)
        fig.subplots_adjust(hspace=0.6, top=1-0.75/figsize[1])
        fig.suptitle(f'{mode}: Fraction of tiny values', fontsize=16)


LayerOutputAnalysisHookCallback = LayerAnalysisCallback


class EarlyStopping(Callback):
    _order = -10000  # 保证最后执行

    def __init__(self, monitor='train', patience=10, min_delta=0., **kwargs):
        """
        Args:
            monitor: train loss or val loss
            patience: max patience epochs of getting worse
            min_delta: 小于 min_delta 的提升被认为没有变化
        """
        super().__init__(**kwargs)
        assert monitor in ['train', 'val'], '"monitor" must be "train" or "val"'
        self.monitor = monitor
        self.patience = patience
        self.min_delta = min_delta
        self.patience_num = 0

    def after_epoch(self):
        loss = self.messages['metric_values_epoch'][self.monitor]['loss']
        if loss > self.best_loss - self.min_delta:
            self.patience_num += 1
        else:
            self.patience_num = 0
        if self.patience_num >= self.patience:
            print('\n ... Early stopping is triggered!\n')
            raise CancelTrainException()


class GradientClipping(Callback):
    def __init__(self, max_norm=0., **kwargs):
        super().__init__(**kwargs)
        self.max_norm = max_norm

    def after_backward(self):
        if self.max_norm:
            nn.utils.clip_grad_norm_(self.model.parameters(), self.max_norm)



if __name__ == '__main__':
    pass
    # a = torch.arange(0, 100)
    # p = torch.linspace(0.01, 1, 100)

    # annealings = "NO LINEAR COS EXP".split()
    # fns = [sched_no, sched_lin, sched_cos, sched_exp]
    # for fn, t in zip(fns, annealings):
    #     f = fn(2, 1e-2)
    #     plt.plot(a, [f(o) for o in p], label=t)
    # plt.legend()
    # plt.show()

    # sched = combine_scheds([0.3, 0.7], [sched_cos(0.3, 0.6), sched_cos(0.6, 0.2)])  # pylint: disable=no-value-for-parameter
    # plt.plot(a, [sched(o) for o in p])
    # plt.show()



"""
class tree of interpreters:
    InterpreterBase ─┬─> Interpreter                                  (for common dataset)
                     ├─> PyGInterpreter                               (for PyG graph dataset)
                     ├─> DGLInterpreter                               (for DGL graph dataset)
                     └─> GraphInterpreter ─┬─> PyGGraphInterpreter    (for PyG dataset with single graph)
                                           └─> DGLGraphInterpreter    (for DGL dataset with single graph)
"""


from .basic_callbacks import Callback
from ..utils import ToolBox as tbox
import torch
import numpy as np
import matplotlib.pyplot as plt
import itertools
import networkx as nx


def test_phase(interp, phase):
    assert phase in ['train', 'val', 'test'], '"phase" must be "train", "val" or "test"'
    bs = getattr(interp, f'_ybs_{phase}')
    if not bs:
        raise Exception(f'!!! "{phase}" is not performed')


class InterpreterBase(Callback):
    _name = 'interpreter'

    def __init__(self, task_type, learner, multi_out):
        """
        Args:
            task_type: type of leaning task
            learner: learner
            multi_out: is the model have multi output
        """
        super().__init__(as_attr=True, learner=learner)
        assert task_type in ['classify', 'regress']  # 分类、回归任务
        self.task_type = task_type
        self.multi_out = multi_out
        self.cpu = torch.device('cpu')

        self._xbs_train = []
        self._xbs_val = []
        self._xbs_test = []
        self._ybs_train = []
        self._ybs_val = []
        self._ybs_test = []
        self._predbs_train = []
        self._predbs_val = []
        self._predbs_test = []

    def begin_fit(self):
        self.epoch_th = 0
        if self.task_type == 'classify':
            self.c_matrix_train = None
            self.c_matrix_val = None
            self.c_matrix_test = None
            self.c_dict_train = None
            self.c_dict_val = None
            self.c_dict_test = None

    def begin_epoch(self):
        self.epoch_th += 1

    def after_pred(self):
        if self.epoch_th == self.epochs or self.state == 'test':  # last epoch or test epoch
            getattr(self, f'_predbs_{self.state}').append(self.predb)
            getattr(self, f'_ybs_{self.state}').append(self.yb)
            getattr(self, f'_xbs_{self.state}').append([x.to(self.cpu) for x in self.xb])  # 放入CPU中避免占用GPU存储

    def cat(self, phase):
        raise Exception("To be rewrited!!!")

    def create_top_indices(self, top_indices, phase):
        batch_size = len(getattr(self, f'_ybs_{phase}')[0])
        top_index = {
            'batch': [i // batch_size for i in top_indices],
            'index': [i % batch_size for i in top_indices]
        }
        return top_index

    def top_data(self, metric, phase='train', largest=True, k=0):
        """
        返回metric指标最大（largest=True）或最小（largest=False）的k个数据。
        Args:
            metric: 计算指标
            phase: 分析对象，'train', 'val' 或 'test'分别表示训练数据、验证数据或测试数据
            largest: 返回最大还是最小的k个数据
        Return: (top_scores, top_xs, top_ys, top_preds, top_index)即
                (metric最大值, 最大值对应的x, 最大值对应的y, 最大值对应的pred, {最大值所在batch及在batch中的位置})
        """
        test_phase(self, phase)

        self.cat(phase)
        scores = []
        with torch.no_grad():
            for preb, yb in zip(getattr(self, f'_predbs_{phase}'), getattr(self, f'_ybs_{phase}')):
                scores.append(metric(preb, yb).detach().cpu())
        scores = torch.cat(scores)
        if k == 0:
            k = len(scores)
        top_k = scores.topk(k, largest=largest)
        top_indices = top_k.indices.numpy()

        top_scores = top_k.values.numpy()
        # top_xs = getattr(self, f'_x_{phase}')[top_indices]
        top_xs = [getattr(self, f'_x_{phase}')[i] for i in top_indices]
        top_ys = getattr(self, f'_y_{phase}')[top_indices].numpy()
        top_preds = getattr(self, f'_pred_{phase}')[top_indices].numpy()

        return top_scores, top_xs, top_ys, top_preds, self.create_top_indices(top_indices, phase)

    def confusion_matrix(self, phase='train', return_dict=False):
        """
        混淆矩阵
        """
        if self.task_type != 'classify':
            raise Exception('Confusion matrix only in "classify" tasks!')
        test_phase(self, phase)
        # 若已存在，不必再计算
        if return_dict:
            c_dict = getattr(self, f'c_dict_{phase}')
            if c_dict is not None:
                return c_dict
        else:
            c_matrix = getattr(self, f'c_matrix_{phase}')
            if c_matrix is not None:
                return c_matrix

        self.cat(phase)

        pred_size = getattr(self, f'_pred_{phase}').size()
        if len(pred_size) == 1:  # for sigmoid output classification
            self.class_num = 2
        else:
            self.class_num = pred_size[1]

        c_dict = {}
        for x, y, pred in zip(getattr(self, f'_x_{phase}'),
                              getattr(self, f'_y_{phase}'),
                              getattr(self, f'_pred_{phase}')):
            if len(pred_size) == 1:
                key = (int(y.numpy().tolist()), pred.round().int().numpy().tolist())
            else:
                key = (int(y.numpy().tolist()), int(pred.argmax().numpy().tolist()))
            if key not in c_dict:
                c_dict[key] = []
            c_dict[key].append(x)

        c_matrix = np.zeros([self.class_num, self.class_num], dtype=int)
        for key, item in c_dict.items():
            c_matrix[key[0], key[1]] = len(item)

        setattr(self, f'c_dict_{phase}', c_dict)
        setattr(self, f'c_matrix_{phase}', c_matrix)

        if return_dict:
            return c_dict
        else:
            return c_matrix

    def plot_confusion(self, phase='train', title='Confusion matrix', class_names=None,
                       normalize=False, norm_dec=2, cmap='Blues', **kwargs):
        """
        画出混淆矩阵
        """
        if self.task_type != 'classify':
            raise Exception('Confusion matrix only in "classify" tasks!')
        test_phase(self, phase)

        c_matrix = self.confusion_matrix(phase)
        data_size = c_matrix.sum()
        if normalize:
            c_matrix = c_matrix.astype('float') / c_matrix.sum(axis=1)[:, np.newaxis]

        fig = plt.figure(**kwargs)

        plt.imshow(c_matrix, interpolation='nearest', cmap=cmap)
        plt.title(f'{title} - {phase}({data_size})')
        if class_names and len(tbox.listify(class_names)) == self.class_num:
            tick_marks = np.arange(self.class_num)
            plt.xticks(tick_marks, class_names, rotation=90)
            plt.yticks(tick_marks, class_names, rotation=0)

        thresh = c_matrix.max() / 2.
        for i, j in itertools.product(range(c_matrix.shape[0]), range(c_matrix.shape[1])):
            coeff = f'{c_matrix[i, j]:.{norm_dec}f}' if normalize else f'{c_matrix[i, j]}'
            plt.text(j, i, coeff, horizontalalignment="center", verticalalignment="center",
                     color="white" if c_matrix[i, j] > thresh else "black")

        ax = fig.gca()
        ax.set_ylim(self.class_num-.5, -.5)

        plt.tight_layout()
        plt.ylabel('Real')
        plt.xlabel('Pred')
        plt.grid(False)
        fig.subplots_adjust(bottom=0.1)
        return fig

    def most_confused(self, phase='train', k=5):
        """
        错分最多的情况
        """
        if self.task_type != 'classify':
            raise Exception('Confusion matrix only in "classify" tasks!')
        test_phase(self, phase)

        c_dict = self.confusion_matrix(phase, return_dict=True)
        c_items = [c for c in c_dict.items() if c[0][0] != c[0][1]]
        return sorted(c_items, key=lambda e: len(e[1]), reverse=True)[:k]


class Interpreter(InterpreterBase):
    def __init__(self, task_type='classify', learner=None, multi_out=False):
        """
        Args:
            task_type: type of leaning task
            learner: learner
            multi_out: is the model have multi output
        """
        super().__init__(task_type, learner, multi_out)

    def cat(self, phase):
        if getattr(self, f'_x_{phase}', None) is None:
            # there may be multiple inputs
            multi_xbs = list(zip(*getattr(self, f'_xbs_{phase}')))
            multi_xbs = map(torch.cat, multi_xbs)
            setattr(self, f'_x_{phase}', list(zip(*multi_xbs)))
        if getattr(self, f'_y_{phase}', None) is None:
            setattr(self, f'_y_{phase}', torch.cat(getattr(self, f'_ybs_{phase}')).detach().cpu())
        if getattr(self, f'_pred_{phase}', None) is None:
            if not self.multi_out:
                setattr(self, f'_pred_{phase}', torch.cat(getattr(self, f'_predbs_{phase}')).detach().cpu())
            else:
                predbs = getattr(self, f'_predbs_{phase}')
                predbs = [torch.cat(predb, 1) for predb in predbs]
                setattr(self, f'_pred_{phase}', torch.cat(predbs).detach().cpu())


def get_size(value, max_value, min_value, max_size, min_size, largest):
    size = 0
    if largest:
        size = (value - min_value)/(max_value - min_value)
    else:
        size = (max_value - value)/(max_value - min_value)
    size = size * (max_size - min_size) + min_size
    return round(size)


class GraphInterpreter(InterpreterBase):
    def __init__(self, task_type='classify', learner=None, multi_out=False):
        """
        Args:
            task_type: type of leaning task
            learner: learner
            multi_out: is the model have multi output
        """
        super().__init__(task_type, learner, multi_out)

    def get_features(self, data):  # pylint: disable=unused-argument
        return None

    def cat(self, phase):
        if getattr(self, f'_x_{phase}', None) is None:
            # batches of y
            ys = getattr(self, f'_ybs_{phase}')[0]
            mask = ys.mask
            ys = ys.data[mask]
            setattr(self, f'_y_{phase}', ys.detach().cpu())
            setattr(self, f'_mask_{phase}', mask)

            # batches of x
            data = getattr(self, f'_xbs_{phase}')[0]
            setattr(self, f'_x_{phase}', self.get_features(data)[mask])
            # setattr(self, f'_x_{phase}', data[1][mask])

            # batches of pred
            if not self.multi_out:
                setattr(self, f'_pred_{phase}', torch.cat(getattr(self, f'_predbs_{phase}'))[mask].detach().cpu())
            else:
                predbs = getattr(self, f'_predbs_{phase}')
                predbs = [torch.cat(predb, 1)[mask] for predb in predbs]
                setattr(self, f'_pred_{phase}', torch.cat(predbs).detach().cpu())

    def create_top_indices(self, top_indices, phase):
        mask = getattr(self, f'_mask_{phase}').numpy()
        index = np.where(mask == True)[0]
        top_index = {
            'index': [index[i] for i in top_indices]
        }
        return top_index

    def data2nxg(self, data):  # pylint: disable=unused-argument
        return None

    def plot_graph(self, metric, k=0, largest=True, phase='train',
                   layout=None, max_node_size=500, min_node_size=100,
                   label_id=False, label_score=True, dec=2, **kwargs):
        """
        plot the PyG Data or dgl.DGLGraph as Networkx graph
        Args:
            metric: top_data metric. 计算指标，与Learner中的指标不同，需要返回每个数据的计算结果，即无需reduction（如均值）
            k: 多少数据
            largest: 返回最大还是最小的数据
            phase: train, val or test
            layout: networkx layout函数
            max_node_size: 最大节点大小
            min_node_size: 最小节点大小
            label_id: id是否出现在节点label中
            label_score: metric的值是否出现在节点label中
            dec: 若metric值的小数位数
        """
        test_phase(self, phase)

        scores, _, _, _, indecies = self.top_data(metric, phase, largest, 0)
        if k == 0:
            k = len(scores)
        scores = scores[:k]
        indecies = indecies['index'][:k]

        data = getattr(self, f'_xbs_{phase}')[0][0]

        min_score, max_score = min([s for s in scores if s > 0]), max(scores)

        node_size = {}
        label_dict = {}
        for s, i in zip(scores, indecies):
            node_size[i] = get_size(s, max_score, min_score, max_node_size, min_node_size, largest)
            if label_score:
                label_dict[i] = '' if s == 0 else f'{s:0.{dec}f}'
                label_dict[i] = f'{i}-{label_dict[i]}' if label_id else label_dict[i]
            else:
                label_dict[i] = i

        g = self.data2nxg(data)  # pylint: disable=assignment-from-none
        if g.number_of_nodes() > 1000:
            print(f'\nALERT! The network is larger than 1000({g.number_of_nodes()}). Drawing may take a long time!\n')

        node_size = [20 if not node_size.get(i) else node_size[i] for i in g.nodes]

        if layout is None:
            layout = nx.random_layout
        plt.figure()
        pos = layout(g)
        nx.draw_networkx_nodes(g, pos=pos, node_size=node_size, **kwargs)
        nx.draw_networkx_edges(g, pos=pos, **kwargs)
        nx.draw_networkx_labels(g, pos, labels=label_dict, **kwargs)

        return g, node_size, label_dict



import torch
import re
import matplotlib.pyplot as plt
from bijou.utils import ToolBox as tbox
from tqdm import tqdm
from pathlib import Path
import os


class CancelTrainException(Exception):
    pass


class CancelEpochException(Exception):
    pass


class CancelBatchException(Exception):
    pass


def camel2snake(name):
    """
    生成callback的名字
    """
    _camel_re1 = re.compile('(.)([A-Z][a-z]+)')
    _camel_re2 = re.compile('([a-z0-9])([A-Z])')

    s1 = re.sub(_camel_re1, r'\1_\2', name)
    return re.sub(_camel_re2, r'\1_\2', s1).lower()


def dictformat(dic):
    return str({k: f'{v:0.6f}' for k, v in dic.items()})


class Callback():
    """
    Callback的基类
    """
    _order = 0
    _name = None

    def __init__(self, as_attr=False, learner=None):
        """
        Args:
            as_attr: 是否将callback对象作为learner的属性
            learner: callback所属的Learner，若为空则后续可调用set_learner设置
        """
        self.as_attr = as_attr
        self.learner = learner

    def set_learner(self, learner):
        self.learner = learner

    def __getattr__(self, k):
        return getattr(self.learner, k)

    @property
    def name(self):
        if self._name is None:
            name = re.sub(r'Callback$', '', self.__class__.__name__)
            self._name = camel2snake(name or 'callback')
        return self._name

    def __call__(self, cb_name):
        f = getattr(self, cb_name, None)
        if f and f():
            return True
        return False


class AvgStats:
    def __init__(self, metrics, state):
        self.metrics, self.state = tbox.listify(metrics), state
        self.metric_names = ['loss'] + [m.__name__ for m in self.metrics]  # 构造metrics 名称

    def reset(self):
        self.tot_loss, self.count = 0., 0
        self.tot_mets = [0.] * len(self.metrics)

    @property
    def all_stats(self):
        return [self.tot_loss.item()] + self.tot_mets

    @property
    def avg_stats(self):
        if self.count > 0:
            return dict(zip(self.metric_names, [o/self.count for o in self.all_stats]))
        else:
            return {}

    def __repr__(self):
        if not self.count:
            return ""
        return f"{self.state}: {self.avg_stats}"

    def accumulate(self, learner):
        bn = len(learner.xb)  # .shape[0]
        self.tot_loss += learner.loss * bn
        self.count += bn
        for i, m in enumerate(self.metrics):
            self.tot_mets[i] += m(learner.predb, learner.yb).item() * bn


class AvgStatsCallback(Callback):
    """
    该类型的回调最后只用一个，如果需要多个指标，则将多个指标组成列表，传到metrics参数即可
    """
    _order = 1

    def __init__(self, metrics, **kwargs):
        super().__init__(**kwargs)
        self.train_stats = AvgStats(metrics, 'train')
        self.val_stats = AvgStats(metrics, 'val')
        self.test_stats = AvgStats(metrics, 'test')

    def begin_fit(self):
        self.learner.messages['metric_values_batch'] = {'train': None, 'val': None, 'test': None}
        self.learner.messages['metric_values_epoch'] = {}

    def begin_epoch(self):
        self.train_stats.reset()
        self.val_stats.reset()

    def after_epoch(self):
        self.learner.messages['metric_values_epoch']['train'] = self.train_stats.avg_stats
        self.learner.messages['metric_values_epoch']['val'] = self.val_stats.avg_stats

        # update best loss
        loss = self.messages['metric_values_epoch']['train']['loss']
        if loss < self.best_loss:
            self.learner.best_loss = loss

    def begin_test(self):
        self.test_stats.reset()

    def after_loss(self):
        stats = getattr(self, f'{self.state}_stats')
        with torch.no_grad():
            stats.accumulate(self.learner)

    def after_batch(self):
        self.learner.messages['metric_values_batch'][self.state] = getattr(self, f'{self.state}_stats').avg_stats


class ProgressBarCallback(Callback):
    _order = 0
    train_message = ''
    val_message = ''
    test_message = ''

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def begin_epoch(self):
        batch_num = len(self.data.train_dl)
        self.pbar = tqdm(total=batch_num,
                         bar_format=f'E {self.epoch + 1:<4} ' + 'B {n_fmt} {l_bar} {rate_fmt} | {postfix[0]}',
                         unit=' batch', postfix=[''])

    def begin_validate(self):
        self.pbar.reset(total=len(self.data.valid_dl))

    def begin_test(self):
        self.pbar = tqdm(total=len(self.test_dl),
                         bar_format=f'TEST   ' + 'B {n_fmt} {l_bar} {rate_fmt} | {postfix[0]}',
                         unit=' batch', postfix=[''])

    def after_batch(self):
        if self.learner.messages['metric_values_batch'][self.state]:
            message = f'{self.state}-'
            message += str(dictformat(self.messages['metric_values_batch'][self.state])).replace("'", '')
            setattr(self, f'{self.state}_message', message)
            self.learner.messages['metric_values_batch'][self.state] = None

        if self.state == 'train':
            self.pbar.postfix[0] = self.train_message
        elif self.state == 'val':
            self.pbar.postfix[0] = self.train_message + '  ' + self.val_message
        elif self.state == 'test':
            self.pbar.postfix[0] = self.test_message
        self.pbar.update()

    def after_epoch(self):
        self.pbar.close()

    def after_test(self):
        self.pbar.close()


class StatesCallback(Callback):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def begin_fit(self):
        self.learner.p_epochs = 0.
        self.learner.n_iter = 0

    def after_batch(self):
        if self.state != 'train':
            return
        self.learner.p_epochs += 1./self.iters
        self.learner.n_iter += 1

    def begin_epoch(self):
        self.learner.p_epochs = self.epoch
        self.model.train()
        self.learner.state = 'train'

    def begin_validate(self):
        self.model.eval()
        self.learner.state = 'val'

    def begin_test(self):
        self.model.eval()
        self.learner.state = 'test'


class Recorder(Callback):
    metric_of_epochs = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def begin_fit(self):
        self.lrs = [[] for _ in self.opt.param_groups]
        self.losses = []

    def after_batch(self):
        if self.state != 'train':
            return
        for pg, lr in zip(self.opt.param_groups, self.lrs):
            lr.append(pg['lr'])
        self.losses.append(self.loss.detach().cpu())

    def after_epoch(self):
        self.metric_of_epochs.append(self.messages['metric_values_epoch'].copy())

    def plot_lr(self, pgid=-1, **kwargs):
        plt.plot(self.lrs[pgid], **kwargs)

    def plot_loss(self, skip_last=0, **kwargs):
        plt.plot(self.losses[:len(self.losses)-skip_last], **kwargs)

    def plot(self):
        """
        plot learning and losses
        """
        fig = plt.figure(figsize=[10, 4])

        plt.subplot(121)
        self.plot_lr()
        plt.xlabel('n iter')
        plt.ylabel('learning rate')

        plt.subplot(122)
        self.plot_loss(c='r')
        plt.xlabel('n iter')
        plt.ylabel('loss')
        fig.subplots_adjust(wspace=0.3, bottom=0.18)

    def plot_lr_loss(self, skip_last=0, pgid=-1):
        losses = [o.item() for o in self.losses]
        lrs = self.lrs[pgid]
        n = len(losses)-skip_last
        plt.xscale('log')
        plt.plot(lrs[:n], losses[:n])

    def plot_metrics(self):
        trains_vals = ld2dl(self.metric_of_epochs)
        trains = ld2dl(trains_vals['train'])
        vals = trains_vals.get('val')  # 可能无验证集
        if vals:
            vals = ld2dl(vals)
        plt.figure(figsize=[len(trains) * 5, 4])
        for i, m in enumerate(trains):
            plt.subplot(1, len(trains), i + 1)
            plt.plot(trains[m], label='train')
            if vals:
                plt.plot(vals[m], label='val')
            plt.title(m, y=-0.2)
            plt.legend()
        plt.subplots_adjust(bottom=0.2)


def ld2dl(ld):
    dl = {}
    for d in ld:
        for k, v in d.items():
            if k not in dl:
                dl[k] = []
            dl[k].append(v)
    return dl


class CudaCallback(Callback):
    _order = 0

    def __init__(self, device, **kwargs):
        super().__init__(**kwargs)
        if device is None:
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.device = device

    def begin_fit(self):
        self.model.to(self.device)

    def begin_batch(self):
        self.learner.xb = self.to_device(self.xb, True)
        self.learner.yb = self.to_device(self.yb, False)

    def begin_predict(self):
        self.model.to(self.device)
        self.learner.predict_data = self.to_device(self.predict_data)
        # self.learner.predict_data = self.predict_data.to(self.device)

    def to_device(self, batch_data, to_list=True):
        """
        to_list: change single data to list or not
        """
        if isinstance(batch_data, (list, tuple)):
            xb = [xs.to(self.device) for xs in batch_data]  # multi-inputs or multi-outputs
        else:
            if to_list:
                xb = [batch_data.to(self.device)]  # single-inputs or multi-outputs
            else:
                xb = batch_data.to(self.device)
        return xb


class Checkpoints(Callback):
    _order = -1

    def __init__(self, epochs=1, path='./checkpoints', skip_worse=False, clear=False, **kwargs):
        """
        Args:
            epochs: save checkpoint each 'epochs' epochs.
            path: path
            skip_worse: skip at worse loss epoch
            clear: whether clear exists checkpoints in path or not
        """
        super().__init__(**kwargs)
        self.per_epochs = epochs
        self.path = path
        self.skip_worse = skip_worse
        self.best_check_loss = float('inf')

        if clear:
            p = Path(path)
            ckpts = p.glob('*.ckpt')
            for ckpt in ckpts:
                os.remove(ckpt)

    def after_epoch(self):
        if (self.epoch+1) % self.per_epochs == 0:
            if self.skip_worse:
                epoch_loss = self.messages['metric_values_epoch']['train']['loss']
                if epoch_loss < self.best_check_loss:
                    self.checkpoint(path=self.path)
                    self.best_check_loss = epoch_loss
            else:
                self.checkpoint(path=self.path)



from torch_geometric.data import DataLoader as DataLoader_pyg
from .data import DataLoaderBase

class PyGDataLoader(DataLoader_pyg, DataLoaderBase):

    def __iter__(self):
        for data in super().__iter__():
            yield data, data.y



from .basic_callbacks import Callback

class BatchTransformXCallback(Callback):
    _order = 2

    def __init__(self, tfm, **kwargs):
        super().__init__(**kwargs)
        self.tfm = tfm

    def begin_batch(self):
        self.learner.xb = self.tfm(self.xb[0])



from .basic_callbacks import *
from .performance import *
from .transforms import *
from .interpreters import *
try:
    from .interpreters_dgl import *
    from .interpreters_pyg import *
except Exception as e:
    pass



import torch
from bijou.utils import rename
import torch.nn.functional as F
import numpy as np

# --- metrics


@rename('acc')
def accuracy(out, yb):
    return (torch.argmax(out, dim=1) == yb).float().mean()


@rename('acc')
def masked_accuracy(pred, target):
    _, pred = pred.max(dim=1)
    correct = pred[target.mask].eq(target.data[target.mask]).sum()
    acc = correct / target.mask.float().sum()
    return acc


@rename('mse')
def masked_mse(pred, target):
    return F.mse_loss(torch.squeeze(pred[target.mask]), target.data[target.mask])


@rename('mae')
def masked_mae(pred, target):
    pred = torch.squeeze(pred[target.mask])
    target = target.data[target.mask]
    return torch.mean(torch.abs(pred - target))


def ordinal(pred, target):
    """
    真实排序下，对应的预测值的序号
    """
    frame = np.array([pred, target]).transpose()
    frame = frame[(-frame[:, 0]).argsort()]
    frame = np.concatenate([frame, np.expand_dims(np.arange(1, len(pred)+1), 1)], 1)
    frame = frame[(-frame[:, 1]).argsort()]
    pred_ids = frame[:, 2]
    return pred_ids


@rename('map')
def MAP(pred, target):
    """
    Mean average precision(MAP)
    """
    n = len(pred)
    target_ids = np.arange(1, len(pred)+1)
    pred_ids = ordinal(pred, target)

    def p_at_n(p_ids, t_ids, n):
        return len(set(p_ids[:n]).intersection(set(t_ids[:n])))/n
    return np.average([p_at_n(pred_ids, target_ids, i) for i in range(1, n+1)])


@rename('map')
def masked_MAP(pred, target):
    """
    Mean average precision(MAP)
    """
    pred = torch.squeeze(pred[target.mask]).detach().cpu().numpy()
    target = target.data[target.mask].detach().cpu().numpy()
    return MAP(pred, target)


def DCG_at_n(pred, target, n):
    """
    Discount Cumulative Gain (DCG@n)
    """
    frame = np.array([pred, target]).transpose()
    frame = frame[(-frame[:, 0]).argsort()]
    frame = frame[:n]
    return np.sum([t/np.log2(i+2) for i, (_, t) in enumerate(frame)])


def NDCG_at_n(n, pred, target):
    """
    Normalized discount cumulative gain (NDCG@n)
    """
    return DCG_at_n(pred, target, n)/DCG_at_n(target, target, n)


def masked_NDCG_at_n(n, pred, target):
    """
    Masked normalized discount cumulative gain (NDCG@n)
    """
    pred = torch.squeeze(pred[target.mask]).detach().cpu().numpy()
    target = target.data[target.mask].detach().cpu().numpy()
    return NDCG_at_n(n, pred, target)


# --- losses

def masked_nll_loss(pred, target):
    return F.nll_loss(pred[target.mask], target.data[target.mask])

def masked_cross_entropy(pred, target):
    return F.cross_entropy(pred[target.mask], target.data[target.mask])

def masked_TOP_loss(pred, target):
    """
    Top One Probability(TOP) loss，from <<Learning to Rank: From Pairwise Approach to Listwise Approach>>
    """
    pred = torch.squeeze(pred[target.mask])
    target = target.data[target.mask]
    pred_p = torch.softmax(pred, 0)
    target_p = torch.softmax(target, 0)
    loss = torch.mean(-torch.sum(target_p*torch.log(pred_p)))
    return loss



import networkx as nx
from .interpreters import InterpreterBase, GraphInterpreter, test_phase
import torch
import matplotlib.pyplot as plt


def pyg_data2g(data):
    if data.is_directed():
        g = nx.DiGraph()
    else:
        g = nx.Graph()
    edge_index = zip(data.edge_index[0].numpy(), data.edge_index[1].numpy())
    g.add_edges_from(edge_index)
    return g


class PyGInterpreter(InterpreterBase):
    def __init__(self, task_type='classify', learner=None, multi_out=False):
        """
        Args:
            task_type: type of leaning task
            learner: learner
            multi_out: is the model have multi output
        """
        super().__init__(task_type, learner, multi_out)

    def cat(self, phase):
        if getattr(self, f'_x_{phase}', None) is None:
            databs = getattr(self, f'_xbs_{phase}')
            data_list = []
            for b in databs:
                data_list.extend(b[0].to_data_list())
            setattr(self, f'_x_{phase}', data_list)
        if getattr(self, f'_y_{phase}', None) is None:
            setattr(self, f'_y_{phase}', torch.cat(getattr(self, f'_ybs_{phase}')).detach().cpu())
        if getattr(self, f'_pred_{phase}', None) is None:
            if not self.multi_out:
                setattr(self, f'_pred_{phase}', torch.cat(getattr(self, f'_predbs_{phase}')).detach().cpu())
            else:
                predbs = getattr(self, f'_predbs_{phase}')
                predbs = [torch.cat(predb, 1) for predb in predbs]
                setattr(self, f'_pred_{phase}', torch.cat(predbs).detach().cpu())

    def plot_graph(self, data, **kwargs):
        g = pyg_data2g(data)
        nx.draw(g, **kwargs)


class PyGGraphInterpreter(GraphInterpreter):
    def get_features(self, data):
        return data[0].x

    def data2nxg(self, data):
        return pyg_data2g(data)



"""
class tree of dataloaders:
    DataLoaderBase ─┬─> DataLoader                            (for common dataset)
                    │        ├─────────> PyGDataLoader        (for PyG graph dataset)
                    │        └─────────> DGLDataLoader        (for DGL graph dataset)
                    └─> GraphLoader                           (for PyG or DGL dataset with single Graph)
"""

from torch.utils.data import DataLoader as TrochDataLoader
import torch


class Dataset():
    def __init__(self, x, y):
        self.x, self.y = x, y

    def __len__(self):
        return len(self.x)

    def __getitem__(self, i):
        return self.x[i], self.y[i]


class DataLoaderBase:
    @classmethod
    def loaders(cls, train_ds, val_ds, test_ds=None, batch_size=1):
        train_dl = cls(train_ds, batch_size=batch_size, shuffle=True)
        val_dl = cls(val_ds, batch_size=batch_size)
        if test_ds is None:
            return train_dl, val_dl
        else:
            return train_dl, val_dl, cls(test_ds, batch_size=batch_size)


class DataLoader(TrochDataLoader, DataLoaderBase):
    def __init__(self, *p, **pd):
        super().__init__(*p, **pd)


class MaskedTensor:
    """
    for semi-supervised learning
    """
    def __init__(self, data, mask):
        self.data = data
        self.mask = mask

    def to(self, device):
        self.data = self.data.to(device)
        self.mask = self.mask.to(device)
        return self

    def __len__(self):
        return len(self.data)


class GraphLoader(DataLoaderBase):
    def __init__(self, graph, labels, mask, features=None):
        """
        Args:
            graph: torch_geometric.data.data.Data或dgl.DGLGraph对象
        """
        if not isinstance(mask, torch.Tensor):
            mask = torch.tensor(mask, dtype=torch.bool)
        if not isinstance(labels, torch.Tensor):
            labels = torch.tensor(labels, dtype=torch.long)

        # GraphLoader must have following three attributes
        if features is None:
            self.data = graph
        else:
            self.data = (graph, features)
        self.label = MaskedTensor(labels, mask)
        self.mask = mask

    @classmethod
    def loaders(cls, g, labels, masks, features=None):
        return [cls(g, labels, mask, features) for mask in masks]

    def __len__(self):
        if torch.sum(self.mask.int()) > 0:    # mask may all False
            return 1
        else:
            return 0

    def __iter__(self):
        yield self.data, self.label


class DataBunch():
    def __init__(self, train_dl, valid_dl=None):
        self.train_dl, self.valid_dl = train_dl, valid_dl

    @classmethod
    def from_dataset(cls, train_ds, valid_ds=None, batch_size=1):
        train_dl = DataLoader(train_ds, batch_size=batch_size)
        if valid_ds is not None:
            valid_dl = DataLoader(valid_dl, batch_size=batch_size)
        else:
            valid_dl = None
        return cls(train_dl, valid_dl)

    @property
    def train_ds(self):
        return self.train_dl.dataset

    @property
    def valid_ds(self):
        if self.valid_dl is None:
            return None
        return self.valid_dl.dataset


class DataProcess:
    """
    数据处理工具函数
    """
    @classmethod
    def normalize(cls, x, m, s):
        """
        标准化
        """
        return (x-m)/s

    @classmethod
    def stats(cls, x):
        """
        均值，标准差
        """
        return x.mean(), x.std()

    @classmethod
    def get_dls(cls, train_ds, valid_ds, bs, **kwargs):
        return (DataLoader(train_ds, batch_size=bs, shuffle=True, **kwargs),
                DataLoader(valid_ds, batch_size=bs*2, **kwargs))

    @classmethod
    def normalize_to(cls, train, valid, test=None):
        m, s = train.mean(), train.std()
        if test is None:
            return cls.normalize(train, m, s), cls.normalize(valid, m, s)
        else:
            return cls.normalize(train, m, s), cls.normalize(valid, m, s), cls.normalize(test, m, s)

    @classmethod
    def flatten(cls, x):
        return x.view(x.shape[0], -1)



"""
后续：
  数据预处理
  各种计算指标
  各种优化技术：参数初始化
"""

import torch
from bijou.data import DataLoaderBase
from bijou.utils import ToolBox as tbox
import bijou.callbacks as cbks
import matplotlib.pyplot as plt
from pathlib import Path
import time


class Learner():
    def __init__(self, model, opt, loss_func, data=None, metrics=None, callbacks=None, device=None):
        """
        Args:
            model: pytorch 模型
            opt: 优化器
            loss_func: 损失函数
            data: DataBunch
            metrics: 性能评价指标或评价指标列表
            callbacks: callbacks对象或类，或者其列表
            device: cpu or gpu device
        """

        self.model, self.opt, self.data, self.loss_func = model, opt, data, loss_func
        self.state = 'train'  # 'train', 'val', 'test'
        self.messages = {}    # 存放需要在不同callbacks之间共享的信息
        self.epoch = 0
        self.best_loss = float('inf')

        cbs = tbox.listify(callbacks)
        # 添加一些必要的回调
        cbs.append(cbks.Recorder(as_attr=True))
        cbs += [cbks.StatesCallback(),
                cbks.CudaCallback(device=device),
                cbks.ProgressBarCallback(),
                cbks.AvgStatsCallback(metrics=metrics)]

        cb_list = []
        for cb in cbs:
            if isinstance(cb, cbks.Callback):
                cb_obj = cb
            else:
                cb_obj = cb()

            if cb_obj.as_attr:
                setattr(self, cb_obj.name, cb_obj)
            cb_obj.set_learner(self)
            cb_list.append(cb_obj)

        self.cbs = cb_list

    def __call__(self, cb_name, reverse=False):
        res = False
        cb_list = sorted(self.cbs, key=lambda x: x._order)  # pylint: disable=protected-access
        if reverse:
            cb_list = reversed(cb_list)
        for cb in cb_list:
            res = cb(cb_name) or res
        return res

    def one_batch(self, xb, yb):
        try:
            self.xb, self.yb = xb, yb
            self('begin_batch', reverse=False)
            self.predb = self.model(*self.xb)  # There may be multiple inputs
            self('after_pred', reverse=True)
            self.loss = self.loss_func(self.predb, self.yb)
            self('after_loss', reverse=True)
            if self.state != 'train':  # 若不是训练状态，则结束batch
                return
            self.loss.backward()
            self('after_backward', reverse=True)
            self.opt.step()
            self('after_step', reverse=True)
            self.opt.zero_grad()
        except cbks.CancelBatchException:
            self('after_cancel_batch', reverse=True)
        finally:
            self('after_batch', reverse=True)

    def all_batches(self, dl):
        self.iters = len(dl)
        try:
            for xb, yb in dl:
                self.one_batch(xb, yb)
        except cbks.CancelEpochException:
            self('after_cancel_epoch', reverse=True)

    def fit(self, epochs):
        self.epochs, self.loss = epochs, torch.tensor(0.)
        start_epoch = self.epoch
        try:
            self('begin_fit', reverse=False)
            for epoch in range(epochs):
                if not self('begin_epoch', reverse=False):
                    self.all_batches(self.data.train_dl)

                if self.data.valid_dl is not None:
                    with torch.no_grad():
                        if not self('begin_validate', reverse=False):
                            self.all_batches(self.data.valid_dl)
                self('after_epoch', reverse=True)
                self.epoch = epoch + start_epoch + 1  # resume from checkpoint

        except cbks.CancelTrainException:
            self('after_cancel_train', reverse=True)
        finally:
            self('after_fit', reverse=True)
            # self.model, self.opt, self.loss_func, self.data = None, None, None, None

    def test(self, test_dl):
        if len(test_dl) == 0:
            return
        self.test_dl = test_dl
        if not self('begin_test', reverse=False):
            with torch.no_grad():
                self.all_batches(test_dl)
        self('after_test', reverse=True)
        self.test_dl = None

    def predict(self, dataset):
        """
        dataset: Dataloader或者可直接计算的数据如Tensor或PyG Data
        """

        if hasattr(dataset, 'to'):  # whether it has the "to(device)" method
            dataset = [dataset]
        is_dl = isinstance(dataset, DataLoaderBase)

        preds = []
        for batch in dataset:
            self.predict_data = batch[0] if is_dl else batch
            self('begin_predict', reverse=False)
            with torch.no_grad():
                preds.append(self.model(*self.predict_data))
        if preds and isinstance(preds[0], tuple):  # multiple output
            return preds # predict result of batches
        else:
            return torch.cat(preds, 0)

    def fit_one_cycle(self, epochs, stage=(0.3, 0.7), start_lr=0.01, high_lr=0.5, end_lr=0.01):
        sched = cbks.combine_scheds(stage,
                                    cbks.cos_1cycle_anneal(start_lr, high_lr, end_lr)
                                    )  # pylint: disable=no-value-for-parameter
        lr_shed_cb = cbks.ParamScheduler('lr', sched, learner=self)
        self.cbs += [lr_shed_cb]
        self.fit(epochs)

        self.cbs.remove(lr_shed_cb)

    def find_lr(self, max_iter=100, min_lr=1e-6, max_lr=10, skip_last=5):
        lr_find_cbk = cbks.LR_Find(max_iter, min_lr, max_lr, learner=self)
        self.cbs += [lr_find_cbk]
        self.fit(max_iter)

        self.recorder.plot_lr_loss(skip_last=skip_last)
        plt.xlabel('learning rate')
        plt.ylabel('loss')

        self.cbs.remove(lr_find_cbk)

    def save_model(self, path='./model.pkl'):
        """
        保存模型
        """
        path = Path(path)
        if not path.parent.exists():
            path.parent.mkdir()
        torch.save(self.model, path)

    @classmethod
    def load_model(cls, path):
        """
        加载模型
        """
        model = torch.load(path)
        return model

    def load_checkpoint(self, check_name=None, check_latest=False, check_best=False,
                        check_opt=True, path='./checkpoints'):
        """
        load learner checkpoint
        Args:
            check_name: checkpoint file name
            check_latest: load latest checkpoint
            check_best: load best loss checkpoint
            check_opt: load opt state or not
            path: checkpoint folder path
        """
        path = Path(path)
        name_dict = {f.name: [float(e) for e in f.name[:-5].split('-')] for f in path.iterdir()
                     if not f.is_dir() and f.name.endswith('.ckpt')}
        # find latest or best checkpoint
        if not check_name:
            if check_name is None and not check_latest and not check_best:
                check_latest = True
            if check_latest:
                check_name = max(name_dict.items(), key=lambda e: e[1][0])[0]
                print('- Loadding the latest checkpoint ...')
            elif check_best:
                check_name = min(name_dict.items(), key=lambda e: e[1][1])[0]
                print('- Loadding the best checkpoint ...')
            if not check_name:
                print('\nNo checkpoint!')
                return
        else:
            if len(check_name) < 25:
                ckn = [f for f in name_dict.keys() if f.startswith(check_name)]
                if ckn:
                    check_name = ckn[0]
            elif not check_name.endswith('.ckpt'):
                check_name = check_name.strip() + '.ckpt'

        check_info = torch.load(path/check_name)

        # load learner state
        self.epoch = 0  # check_info['epoch'] + 1
        self.best_loss = check_info['best_loss']

        # load mode state
        model_dict = self.model.state_dict()
        update_state = {k: v for k, v in check_info['model_state'].items() if k in model_dict.keys()}
        model_dict.update(update_state)
        self.model.load_state_dict(model_dict)

        # load opt sate
        if check_opt:
            self.opt.load_state_dict(check_info['opt_state'])

    def checkpoint(self, path='./checkpoints'):
        """
        save learner checkpoint
        """
        path = Path(path)
        if not path.exists():
            path.mkdir()
        check_info = {
            'epoch': self.epoch,
            'best_loss': self.best_loss,
            'model_state': self.model.state_dict(),
            'opt_state': self.opt.state_dict(),
        }
        check_name = f'{int(time.time())}-{self.best_loss:0.10f}-{self.epoch+1}'
        torch.save(check_info, path/f'{check_name}.ckpt')



from bijou import __version__
import setuptools

desc = 'A fastai-like framework for training, tuning and probing pytorch models, which is compatible with pytorch_geometric.'

with open("README.md", 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="bijou",
    version=__version__,
    author="hitlic",
    author_email="liuchen.lic@gmail.com",
    license='MIT',
    description=desc,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/hitlic/bijou",
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    keywords='',
    packages=setuptools.find_packages(exclude=['examples', 'datasets',]),
    py_modules=[],
    install_requires=['torch > 1.1', 'tqdm > 4.40', 'matplotlib > 3.1', 'networkx >= 2.3', 'requests > 2.20.0'],
    python_requires='>=3.5'
)



import torch
import pickle
import gzip
from pathlib import Path
import requests
from tqdm import tqdm
import zipfile
from math import ceil

def download(url, path, fname):
    print(f"Downloading from {url}")
    resp = requests.get(url=url, stream=True)
    total = ceil(int(resp.headers['Content-Length'])/1024)
    with open(path/fname, "wb") as f:
        for data in tqdm(iterable=resp.iter_content(1024), total=total, unit='k'):
            f.write(data)
    zf = zipfile.ZipFile(path/fname)
    zf.extractall(path=path)
    (path/fname).unlink()


def mnist(path='./datasets'):
    path = Path(path)
    if not (path).exists():
        path.mkdir(parents=True)
    if not (path/'mnist').exists():
        url = 'https://github.com/hitlic/bijou/raw/master/datasets/mnist.zip'
        download(url, path, 'mnist.zip')

    path = path/'mnist'/'mnist.pkl.gz'
    with gzip.open(path, 'rb') as f:
        ((x_train, y_train), (x_valid, y_valid), (x_test, y_test)) = pickle.load(f, encoding='latin-1')
    return map(torch.tensor, (x_train, y_train, x_valid, y_valid, x_test, y_test))


def pyg_cora(path='./datasets'):
    '''
    https://people.cs.umass.edu/~mccallum/data/
    '''
    path = Path(path)
    if not (path).exists():
        path.mkdir(parents=True)
    if not (path/'PyG-Cora').exists():
        url = 'https://github.com/hitlic/bijou/raw/master/datasets/PyG-Cora.zip'
        download(url, path, 'PyG-Cora.zip')
    return path/'PyG-Cora'


def pyg_yoochoose_10k(path='./datasets'):
    '''
    https://2015.recsyschallenge.com/challenge.html
    '''
    path = Path(path)
    if not (path).exists():
        path.mkdir(parents=True)
    if not (path/'PyG-yoochoose_10k').exists():
        url = 'https://github.com/hitlic/bijou/raw/master/datasets/PyG-yoochoose_10k.zip'
        download(url, path, 'PyG-yoochoose_10k.zip')
    return path/'PyG-yoochoose_10k'


def dgl_cora(path='./datasets'):
    '''
    https://people.cs.umass.edu/~mccallum/data/
    '''
    path = Path(path)
    if not (path).exists():
        path.mkdir(parents=True)
    if not (path/'DGL-Cora').exists():
        url = 'https://github.com/hitlic/bijou/raw/master/datasets/DGL-Cora.zip'
        download(url, path, 'DGL-Cora.zip')
    return path/'DGL-Cora'



from .data import DataLoaderBase
import torch
from torch.utils.data import DataLoader as TorchDataLoader
import dgl

def collate(samples):
    # The input `samples` is a list of pairs  (graph, label).
    graphs, labels = map(list, zip(*samples))
    batched_graph = dgl.batch(graphs)
    return batched_graph, torch.tensor(labels)


class DGLDataLoader(TorchDataLoader, DataLoaderBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, collate_fn=collate, **kwargs)



from .interpreters import InterpreterBase, GraphInterpreter
import torch
import networkx as nx
import dgl


class DGLInterpreter(InterpreterBase):
    def __init__(self, task_type='classify', learner=None, multi_out=False):
        """
        Args:
            task_type: type of leaning task
            learner: learner
            multi_out: is the model have multi output
        """
        super().__init__(task_type, learner, multi_out)

    def cat(self, phase):
        if getattr(self, f'_x_{phase}', None) is None:
            databs = getattr(self, f'_xbs_{phase}')
            data_list = []
            for b in databs:
                data_list.extend(dgl.unbatch(b[0]))
            setattr(self, f'_x_{phase}', data_list)
        if getattr(self, f'_y_{phase}', None) is None:
            setattr(self, f'_y_{phase}', torch.cat(getattr(self, f'_ybs_{phase}')).detach().cpu())
        if getattr(self, f'_pred_{phase}', None) is None:
            if not self.multi_out:
                setattr(self, f'_pred_{phase}', torch.cat(getattr(self, f'_predbs_{phase}')).detach().cpu())
            else:
                predbs = getattr(self, f'_predbs_{phase}')
                predbs = [torch.cat(predb, 1) for predb in predbs]
                setattr(self, f'_pred_{phase}', torch.cat(predbs).detach().cpu())

    def plot_graph(self, data, directed=False, **kwargs):
        if directed:
            g = data.to_networkx()
        else:
            g = data.to_networkx().to_undirected()
        nx.draw(g, **kwargs)



class DGLGraphInterpreter(GraphInterpreter):
    def get_features(self, data):
        return data[1]

    def data2nxg(self, data):
        return data.to_networkx()
